"""
Central Authentication Service (CAS) client
"""
import asyncio
import dataclasses
import logging
import urllib.parse
import urllib.request
import xml.etree.ElementTree
from collections.abc import Awaitable, Callable
from typing import Optional, Union, cast

logger = logging.getLogger(__name__)

CAS_NS = {"cas": "http://www.yale.edu/tp/cas"}
CAS_VALIDATE_ENCODING = "utf-8"
CAS_VALIDATE_TIMEOUT = 10.0

HTTPGetFunc = Callable[[str, float], str]
AsyncHTTPGetFunc = Callable[[str, float], Awaitable[str]]


class CASError(Exception):
    def __init__(self, error_code: str, *args: Optional[str]) -> None:
        super().__init__(error_code, *args)
        self.error_code = error_code


class CASInvalidServiceError(CASError):
    def __init__(self, *args: Optional[str]) -> None:
        super().__init__("INVALID_SERVICE", *args)


class CASInvalidTicketError(CASError):
    def __init__(self, *args: Optional[str]) -> None:
        super().__init__("INVALID_TICKET", *args)


@dataclasses.dataclass
class CASUser:
    userid: str
    attributes: dict[str, Optional[str]] = dataclasses.field(default_factory=dict)

    def asdict(self) -> dict[str, Union[str, dict[str, str]]]:
        return dataclasses.asdict(self)


class BaseCASClient:
    def __init__(self, login_url: str, logout_url: str, validate_url: str) -> None:
        self.login_url = login_url
        self.logout_url = logout_url
        self.validate_url = validate_url

    def build_login_url(
        self,
        service: str,
        *,
        callback_post: bool = False,
        **kwargs: str,
    ) -> str:
        params = {"service": service, **kwargs}
        if callback_post:
            params["method"] = "POST"
        qs = urllib.parse.urlencode(params)
        return f"{self.login_url}?{qs}"

    def build_logout_url(self, service: Optional[str] = None, **kwargs: str) -> str:
        if service is None:
            if not kwargs:
                return self.logout_url
            params = kwargs
        else:
            params = {"service": service, **kwargs}
        qs = urllib.parse.urlencode(params)
        return f"{self.logout_url}?{qs}"

    def build_validate_url(self, service: str, ticket: str, **kwargs: str) -> str:
        params = {"service": service, "ticket": ticket, **kwargs}
        qs = urllib.parse.urlencode(params)
        return f"{self.validate_url}?{qs}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"login_url={self.login_url!r}, "
            f"logout_url={self.logout_url!r}, "
            f"validate_url={self.validate_url!r}"
            ")"
        )


class CASClient(BaseCASClient):
    def __init__(
        self,
        login_url: str,
        logout_url: str,
        validate_url: str,
        http_get_func: Optional[HTTPGetFunc] = None,
    ) -> None:
        super().__init__(login_url, logout_url, validate_url)
        if http_get_func is None:
            self._http_get = cast(HTTPGetFunc, _http_get)
        else:
            self._http_get = http_get_func

    @classmethod
    def from_base_url(
        cls,
        base_url: str,
        *,
        login_path: str = "/login",
        logout_path: str = "/logout",
        validate_path: str = "/p3/serviceValidate",
        http_get_func: Optional[HTTPGetFunc] = None,
    ) -> "CASClient":
        return cls(
            login_url=urllib.parse.urljoin(base_url, login_path),
            logout_url=urllib.parse.urljoin(base_url, logout_path),
            validate_url=urllib.parse.urljoin(base_url, validate_path),
            http_get_func=http_get_func,
        )

    def validate(
        self,
        service_url: str,
        ticket: str,
        *,
        timeout: Optional[float] = None,
        **kwargs: str,
    ) -> CASUser:
        if timeout is None:
            timeout = CAS_VALIDATE_TIMEOUT
        target_validate = self.build_validate_url(service_url, ticket, **kwargs)
        logger.debug("Validating %s", target_validate)
        try:
            resp_text = self._http_get(target_validate, timeout)
        except Exception as exc:
            raise CASError(repr(exc)) from exc
        else:
            logger.debug("Response:\n%s", resp_text)
            return parse_cas_response(resp_text)


class AsyncCASClient(BaseCASClient):
    def __init__(
        self,
        login_url: str,
        logout_url: str,
        validate_url: str,
        http_get_func: Optional[AsyncHTTPGetFunc] = None,
    ) -> None:
        super().__init__(login_url, logout_url, validate_url)
        if http_get_func is None:
            self._http_get = cast(AsyncHTTPGetFunc, _async_http_get)
        else:
            self._http_get = http_get_func

    @classmethod
    def from_base_url(
        cls,
        base_url: str,
        *,
        login_path: str = "/login",
        logout_path: str = "/logout",
        validate_path: str = "/p3/serviceValidate",
        http_get_func: Optional[AsyncHTTPGetFunc] = None,
    ) -> "AsyncCASClient":
        return cls(
            login_url=urllib.parse.urljoin(base_url, login_path),
            logout_url=urllib.parse.urljoin(base_url, logout_path),
            validate_url=urllib.parse.urljoin(base_url, validate_path),
            http_get_func=http_get_func,
        )

    async def validate(
        self,
        service_url: str,
        ticket: str,
        *,
        timeout: Optional[float] = None,
        **kwargs: str,
    ) -> CASUser:
        if timeout is None:
            timeout = CAS_VALIDATE_TIMEOUT
        target_validate = self.build_validate_url(service_url, ticket, **kwargs)
        logger.debug("Validating %s", target_validate)
        try:
            resp_text = await self._http_get(target_validate, timeout)
        except Exception as exc:
            raise CASError(repr(exc)) from exc
        else:
            logger.debug("Response:\n%s", resp_text)
            return parse_cas_response(resp_text)


def parse_cas_response(cas_response: str) -> CASUser:
    try:
        root = xml.etree.ElementTree.fromstring(cas_response)
    except Exception as exc:
        raise CASError("INVALID_RESPONSE", repr(exc)) from exc
    else:
        return parse_cas_xml(root)


def parse_cas_xml(root: xml.etree.ElementTree.Element) -> CASUser:
    user_elem = root.find("cas:authenticationSuccess/cas:user", CAS_NS)
    if user_elem is not None:
        attr_elem = root.find("cas:authenticationSuccess/cas:attributes", CAS_NS)
        return parse_cas_xml_user(user_elem, attr_elem)
    raise parse_cas_xml_error(root)


def parse_cas_xml_user(
    user_elem: xml.etree.ElementTree.Element,
    attr_elem: Optional[xml.etree.ElementTree.Element],
) -> CASUser:
    if user_elem.text is None:
        raise CASError("USERNAME_NOT_IN_RESPONSE")
    cas_user = CASUser(userid=user_elem.text)
    if attr_elem is not None:
        tag_ns = "{" + CAS_NS["cas"] + "}"
        for e in attr_elem:
            attr_name = e.tag.replace(tag_ns, "", 1)
            cas_user.attributes[attr_name] = e.text
    return cas_user


def parse_cas_xml_error(root: xml.etree.ElementTree.Element) -> CASError:
    error_code = "Unknown"
    error_elem = root.find("cas:authenticationFailure", CAS_NS)
    if error_elem is not None:
        error_code = error_elem.attrib.get("code", error_code)
        error_text = error_elem.text
        if error_code == "INVALID_TICKET":
            return CASInvalidTicketError(error_text)
        if error_code == "INVALID_SERVICE":
            return CASInvalidServiceError(error_text)
    return CASError(error_code)


def _http_get(url: str, timeout: float = 10.0) -> str:
    with urllib.request.urlopen(url, timeout=timeout) as f:
        data = cast(bytes, f.read())
        return data.decode(CAS_VALIDATE_ENCODING)


async def _async_http_get(url: str, timeout: float = 10.0) -> str:
    fut = asyncio.to_thread(_http_get, url, timeout)
    return await asyncio.wait_for(fut, timeout)



"""
Simple Apereo Central Authentication Service (CAS) client
"""
import logging as _logging

from .core import (
    AsyncCASClient,
    CASClient,
    CASError,
    CASInvalidServiceError,
    CASInvalidTicketError,
    CASUser,
)

__version__ = "0.0.7"

__all__ = [
    "AsyncCASClient",
    "CASClient",
    "CASUser",
    "CASError",
    "CASInvalidServiceError",
    "CASInvalidTicketError",
]

_logging.getLogger(__name__).addHandler(_logging.NullHandler())



import unittest

from cassyy.core import (
    CASClient,
    CASError,
    CASInvalidServiceError,
    CASInvalidTicketError,
    parse_cas_response,
)


class CASClientTestCase(unittest.TestCase):
    cas_login_url = "https://cas.local/login"
    cas_logout_url = "https://cas.local/logout"
    cas_validate_url = "https://cas.local/p3/serviceValidate"
    test_service_url = "https://foo.org"

    def setUp(self):
        self.client = CASClient(
            self.cas_login_url, self.cas_logout_url, self.cas_validate_url
        )

    def test_from_base_url(self):
        c = CASClient.from_base_url("https://cas.local/")
        self.assertEqual(c.login_url, "https://cas.local/login")
        self.assertEqual(c.logout_url, "https://cas.local/logout")
        self.assertEqual(c.validate_url, "https://cas.local/p3/serviceValidate")

    def test_from_base_url_with_alt_paths(self):
        c = CASClient.from_base_url(
            "https://cas.local/",
            login_path="foo",
            logout_path="bar/baz",
            validate_path="/qux",
        )
        self.assertEqual(c.login_url, "https://cas.local/foo")
        self.assertEqual(c.logout_url, "https://cas.local/bar/baz")
        self.assertEqual(c.validate_url, "https://cas.local/qux")

    def test_parse_userid(self):
        s = """
        <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
            <cas:authenticationSuccess>
                <cas:user>jdoe</cas:user>
            </cas:authenticationSuccess>
        </cas:serviceResponse>
        """
        self.assertEqual("jdoe", parse_cas_response(s).userid)

    def test_parse_non_xml(self):
        s = "jdoe"
        with self.assertRaises(CASError) as cm:
            parse_cas_response(s)
        self.assertEqual("INVALID_RESPONSE", cm.exception.error_code)
        self.assertEqual(
            "ParseError('syntax error: line 1, column 0')", str(cm.exception.args[1])
        )

    def test_parse_invalid_ticket(self):
        s = """
        <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
            <cas:authenticationFailure code="INVALID_TICKET">
                Ticket &#39;ST-foo&#39; not recognized
            </cas:authenticationFailure>
        </cas:serviceResponse>
        """
        with self.assertRaises(CASInvalidTicketError) as cm:
            parse_cas_response(s)
        self.assertEqual("INVALID_TICKET", cm.exception.error_code)

    def test_parse_invalid_service(self):
        s = """
        <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
            <cas:authenticationFailure code="INVALID_SERVICE">
                Ticket &#39;ST-338345-KTQdtsv9b5WKtfVfrahU-cas3&#39; does not match supplied service. '
                The original service was &#39;https://foo.org&#39; and the supplied service was &#39;https://foo2.org&#39;.
            </cas:authenticationFailure>
        </cas:serviceResponse>
        """
        with self.assertRaises(CASInvalidServiceError) as cm:
            parse_cas_response(s)
        self.assertEqual("INVALID_SERVICE", cm.exception.error_code)

    def test_build_login_url(self):
        url = self.client.build_login_url(self.test_service_url)
        self.assertEqual(f"{self.cas_login_url}?service=https%3A%2F%2Ffoo.org", url)

    def test_build_login_url_with_postback(self):
        url = self.client.build_login_url(self.test_service_url, callback_post=True)
        self.assertEqual(
            f"{self.cas_login_url}?service=https%3A%2F%2Ffoo.org&method=POST", url
        )

    def test_build_validate_url(self):
        url = self.client.build_validate_url(self.test_service_url, "tix")
        self.assertEqual(
            f"{self.cas_validate_url}?service=https%3A%2F%2Ffoo.org&ticket=tix", url
        )

    def test_build_logout_url(self):
        url = self.client.build_logout_url(self.test_service_url)
        self.assertEqual(f"{self.cas_logout_url}?service=https%3A%2F%2Ffoo.org", url)

    def test_parse_attributes(self):
        s = """
            <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
                <cas:authenticationSuccess>
                    <cas:user>jdoe</cas:user>
                    <cas:attributes>
                        <cas:clientIpAddress>10.0.0.2</cas:clientIpAddress>
                        <cas:isFromNewLogin>true</cas:isFromNewLogin>
                        <cas:mail>jdoe@foo.org</cas:mail>
                        <cas:authenticationDate>2022-01-21T23:03:05.920747Z</cas:authenticationDate>
                        <cas:bypassMultifactorAuthentication>false</cas:bypassMultifactorAuthentication>
                        <cas:authnContextClass>mfa-example</cas:authnContextClass>
                        <cas:successfulAuthenticationHandlers>DuoSecurityAuthenticationHandler</cas:successfulAuthenticationHandlers>
                        <cas:userAgent>Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36</cas:userAgent>
                        <cas:cn>Jane Doe</cas:cn>
                        <cas:credentialType>DuoSecurityCredential</cas:credentialType>
                        <cas:authenticationMethod>DuoSecurityAuthenticationHandler</cas:authenticationMethod>
                        <cas:serverIpAddress>10.0.0.1</cas:serverIpAddress>
                        <cas:longTermAuthenticationRequestTokenUsed>false</cas:longTermAuthenticationRequestTokenUsed>
                        </cas:attributes>
                </cas:authenticationSuccess>
            </cas:serviceResponse>
        """
        cas_user = parse_cas_response(s)
        self.assertEqual("jdoe", cas_user.userid)
        self.assertEqual("jdoe@foo.org", cas_user.attributes["mail"])
        self.assertEqual("Jane Doe", cas_user.attributes["cn"])
        self.assertEqual("10.0.0.2", cas_user.attributes["clientIpAddress"])
